from ast import literal_eval
import logging
import os
import sys
import subprocess

__all__ = [
    'get_dependencies',
    'install_requirements',
]

logger = logging.getLogger(__name__)


def get_dependencies(module, addons_path=None, deps=None):
    """Get all the dependencies of a module without database

    Using `__terp__.py` files and is used to check requirements.txt in the
    dependencies.

    :param module: Module to find the dependencies
    :param addons_path: Path to find the modules
    :return: a listt of dependencies.
    """
    if deps is None:
        deps = []
    pj = os.path.join
    module_path = pj(addons_path, module)
    if not os.path.exists(module_path):
        raise Exception('Module \'{}\' not found in {}'.format(
            module, addons_path
        ))
    terp_path = pj(module_path, '__terp__.py')
    if not os.path.exists(terp_path):
        raise Exception(
            'Module {} is not a valid module. Missing __terp__.py file'.format(
                module
            )
        )
    with open(terp_path, 'r') as terp_file:
        terp = literal_eval(terp_file.read())

    for dep in terp['depends']:
        if dep not in deps:
            deps.append(dep)
            deps += get_dependencies(dep, addons_path, deps)

    return list(set(deps))


def install_requirements(module, addons_path, silent=False, done=None):
    """Install module requirements and its dependecies
    """
    if done is None:
        done = []
    subprocess_kwargs = {}
    if silent:
        FNULL = open(os.devnull, 'w')
        subprocess_kwargs = {
            'stderr': FNULL,
            'stdout': FNULL
        }
    pip = os.path.join(sys.prefix, 'bin', 'pip')
    if os.path.exists(pip):
        modules_requirements = get_dependencies(module, addons_path)
        modules_requirements.append(module)
        for module_requirements in modules_requirements:
            if module_requirements in done:
                continue
            addons_path_module = os.path.join(addons_path, module_requirements)
            req = os.path.join(addons_path_module, 'requirements.txt')
            if os.path.exists(req):
                logger.info('Requirements file %s found. Installing...', req)
                subprocess.check_call([pip, "install", "-r", req], **subprocess_kwargs)
        return modules_requirements
    return [module]



# coding=utf-8
from __future__ import absolute_import
import click
from tqdm import tqdm
import psycopg2
import ConfigParser
from osconf import config_from_environment


@click.group()
@click.option('--config', required=False, default=None)
@click.pass_context
def oopgrade(ctx, config):
    ctx.obj = {}
    if config:
        conf = ConfigParser.ConfigParser()
        conf.read(config)
        ctx.obj = dict(conf.items('options'))
    envconf = config_from_environment('OPENERP')
    ctx.obj.update(envconf)


@oopgrade.group()
def requirements():
    pass


@requirements.command()
@click.pass_obj
def install(conf):
    from oopgrade.oopgrade import get_installed_modules
    from oopgrade.utils import install_requirements
    conn = psycopg2.connect(
        dbname=conf['db_name'], user=conf['db_user'],
        password=conf['db_password'], host=conf['db_host']
    )
    click.echo('Getting installed modules...')
    with conn:
        with conn.cursor() as cursor:
            modules = get_installed_modules(cursor)
    conn.close()
    done = []
    for module in tqdm(modules, desc='Installing'):
        if module in done:
            continue
        done += install_requirements(
            module, conf['addons_path'], silent=True, done=done
        )



# coding=utf-8
from __future__ import absolute_import

from collections import namedtuple
from ast import literal_eval

from lxml import objectify
from oopgrade.oopgrade import get_foreign_keys, logger
from ooquery import OOQuery
from sql import Table


DataRecord = namedtuple('DataRecord', ['id', 'model', 'noupdate', 'vals'])


class DataMigration(object):
    """Data Migration class

    :param content: XML Content to migrate
    :param cursor: Database cursor
    :param module: OpenObject module name
    :param search_params: Dict where key is the model and value is the list
    of fields to do the match.

    Example::

        from oopgrade import DataMigration

        dm = DataMigration(xml_content, cursor, 'module_name', search_params={
            'test.model': ['field1', 'field2']
        })
        dm.migrate()

    In this case when a record for model `test.model` is found it will use the
    fields `field1` and `field2` to do the match it will construct a search
    query as::

        [
            ('field1', '=', 'content_xml_record_field1'),
            ('field2', '=', 'content_xml_record_field2')
        ]

    .. note:: If no search_params is passed **all** the fields from the xml will
              be used to create the search params
    """
    def __init__(self, content, cursor, module, search_params=None):
        self.content = content
        self.cursor = cursor
        self.module = module
        if search_params is None:
            search_params = {}
        self.search_params = search_params.copy()
        self.records = []

    def _record(self, record):
        vals = {}
        noupdate = bool(
            literal_eval(record.getparent().attrib.get('noupdate', '0'))
        )
        for field in record.iter(tag='field'):
            key = field.attrib['name']
            attrs = field.attrib
            if attrs.get('eval'):
                value = literal_eval(attrs['eval'])
            elif attrs.get('ref'):
                value = self._ref(attrs['ref'])
            elif attrs.get('search') and attrs.get('model'):
                value = self._search(attrs['model'], attrs['search'])
            else:
                value = field.text
            vals[key] = value
        return DataRecord(
            record.attrib['id'], record.attrib['model'], noupdate, vals
        )

    def _ref(self, ref):
        if '.' in ref:
            module, xml_id = ref.split('.')
        else:
            xml_id = ref
            module = self.module

        t = Table('ir_model_data')
        select = t.select(t.res_id)
        select.where = (t.module == module) & (t.name == xml_id)

        self.cursor.execute(*select)
        res = self.cursor.fetchone()
        if not res:
            raise KeyError('Reference: {}.{} not found'.format(
                module, xml_id
            ))
        return res[0]

    def _search(self, model, search):
        table = model.replace('.', '_')
        search_params = literal_eval(search)
        q = OOQuery(table, lambda t: get_foreign_keys(self.cursor, t))
        sql = q.select(['id']).where(search_params)
        self.cursor.execute(*sql)
        return self.cursor.fetchone()[0]

    def migrate(self):
        obj = objectify.fromstring(self.content)
        t = Table('ir_model_data')
        for xml_record in obj.iter(tag='record'):
            record = self._record(xml_record)
            self.records.append(record)
            sp = []
            for field in self.search_params.get(record.model, record.vals.keys()):
                sp.append((field, '=', record.vals[field]))
            logger.info('Trying to find existing record with query: {}'.format(
                sp
            ))
            table = record.model.replace('.', '_')
            q = OOQuery(table)
            sql = q.select(['id']).where(sp)
            logger.debug(tuple(sql))
            self.cursor.execute(*sql)
            res_id = self.cursor.fetchone()
            if res_id:
                res_id = res_id[0]
                logger.info('Record {}.{} found! ({} id:{})'.format(
                    self.module, record.id, record.model, res_id
                ))
            else:
                logger.info('Record {}.{} not found!'.format(
                    self.module, record.id
                ))
                # We have to create the model
                table_model = Table(record.model.replace('.', '_'))
                columns = []
                values = []
                for col, value in record.vals.items():
                    columns.append(getattr(table_model, col))
                    values.append(value)

                sql = table_model.insert(
                    columns=columns, values=[values], returning=[table_model.id]
                )
                logger.debug(tuple(sql))
                self.cursor.execute(*sql)
                res_id = self.cursor.fetchone()[0]
                logger.info('Creating record {}.{} ({} id:{})'.format(
                    self.module, record.id, record.model, res_id
                ))

            sql = t.insert(
                columns=[t.name, t.model, t.noupdate, t.res_id, t.module],
                values=[(record.id, record.model, record.noupdate, res_id,
                         self.module)]
            )
            logger.debug(tuple(sql))
            logger.info('Linking model data {}.{} -> record {} id:{}'.format(
                self.module, record.id, record.model, res_id
            ))
            self.cursor.execute(*sql)



# coding=utf-8
from spec.fixtures import get_fixture
from expects import *
from mock import Mock, call

from oopgrade import DataMigration
from oopgrade.data import DataRecord


with description('Migrating _data.xml'):
    with before.all:
        with open(get_fixture('migration_data.xml'), 'r') as f:
            self.xml = f.read()

    with it('must parse xml files with records'):
        cursor = Mock()
        cursor.fetchone.return_value = [435]
        dm = DataMigration(self.xml, cursor, 'module')
        expect(dm.records).to(have_len(0))
        dm.migrate()
        expect(dm.records).to(have_len(4))

    with context('a DataRecord class'):
        with before.all:
            cursor = Mock()
            cursor.fetchone.return_value = [435]
            self.r = DataRecord('id', 'model', 'noupdate', 'vals')
            self.dm = DataMigration(self.xml, cursor, 'module')
            self.dm.migrate()

        with it('must have a id, model and vals properties'):
            expect(self.r).to(have_properties('id', 'model', 'vals'))

        with it('must work without noupdate'):
            record = self.dm.records[0]
            expect(record.noupdate).to(equal(False))

        with it('must work without noupdate=1'):
            record = self.dm.records[2]
            expect(record.noupdate).to(equal(True))

        with it('must add fields to the records'):
            expect(self.dm.records[0].vals.keys()).to(contain('name', 'description'))

        with it('must to permit create a search_params with record fields'):
            search_params = {
                'test.search.model': ['code']
            }
            cursor = Mock()
            cursor.fetchone.return_value = [435]
            dm = DataMigration(self.xml, cursor, 'module', search_params)
            expect(dm.search_params).to(equal(search_params))

        with it('must work with eval attribute'):
            record = self.dm.records[1]
            expect(record.vals['flag']).to(equal(0))

        with context('working with the ref attribute'):

            with it('must work with other module ref attribute'):
                sql = self.dm.cursor.execute.call_args_list[2]
                expected_sql = call(
                    'SELECT "a"."res_id" FROM "ir_model_data" AS "a" '
                    'WHERE (("a"."module" = %s) AND ("a"."name" = %s))',
                    ('other_module', 'xml_id')
                )
                expect(sql).to(equal(expected_sql))

                record = self.dm.records[1]
                expect(record.vals['relation']).to(equal(435))

            with it('must work with internal refs'):
                sql = self.dm.cursor.execute.call_args_list[8]

                expected_sql = call(
                    'SELECT "a"."res_id" FROM "ir_model_data" AS "a" '
                    'WHERE (("a"."module" = %s) AND ("a"."name" = %s))',
                    ('module', 'record_id_0002')
                )
                expect(sql).to(equal(expected_sql))
                record = self.dm.records[3]
                expect(record.vals['test_model_id']).to(equal(435))

        with context('If reference does not exist'):
            with it('must raise a KeyError exception'):

                def callback():
                    xml = """<?xml version="1.0" encoding="UTF-8" ?>
<openerp>
    <data>
        <record id="record_id_0001" model="test.search.model">
            <field name="code">code</field>
            <field name="relation" ref="other_module.xml_id"/>
        </record>
    </data>
</openerp>
"""
                    cursor = Mock()
                    cursor.fetchone.return_value = []
                    dm = DataMigration(xml, cursor, 'module')
                    dm.migrate()

                expect(callback).to(raise_error(
                    KeyError,
                    'Reference: other_module.xml_id not found'
                ))

        with description('Working with search attribute'):
            with it('must work with search attribute'):
                cursor = Mock()
                cursor.fetchone.side_effect = (
                    [1],
                    [6],
                    [123],
                    [3],
                    [2],
                    [2],
                    [],
                    [4]
                )
                dm = DataMigration(self.xml, cursor, 'module')
                dm.migrate()
                record = dm.records[1]
                expect(record.vals['partner_id']).to(equal(123))

    with description('Migrating'):
        with it('must create the records into ir_model_data'):
            cursor = Mock()
            cursor.fetchone.side_effect = (
                [],  # No record record_id_0001
                [1],  # Creating record_id_0001
                [11],
                [123],
                [3],    # record_id_0003 found
                [2],    # record_id_0002 found
                [2],
                [],
                [4]
            )
            dm = DataMigration(self.xml, cursor, 'module', search_params={
                'test.search.model': ['code']
            })
            dm.migrate()
            expected_sql = [
                call('SELECT "a"."id" FROM "test_model" AS "a" WHERE (("a"."name" = %s) AND ("a"."description" = %s))', ('name', 'this is a description')),
                call('INSERT INTO "test_model" ("name", "description") VALUES (%s, %s) RETURNING "id"', ('name', 'this is a description')),
                call('INSERT INTO "ir_model_data" ("name", "model", "noupdate", "res_id", "module") VALUES (%s, %s, %s, %s, %s)', ('record_id_0001', 'test.model', 0, 1, 'module')),
                call('SELECT "a"."res_id" FROM "ir_model_data" AS "a" WHERE (("a"."module" = %s) AND ("a"."name" = %s))', ('other_module', 'xml_id')),
                call('SELECT "a"."id" FROM "res_partner" AS "a" WHERE (("a"."ref" = %s))', ('123',)),
                call('SELECT "a"."id" FROM "test_search_model" AS "a" WHERE (("a"."code" = %s))', ('code',)),
                call('INSERT INTO "ir_model_data" ("name", "model", "noupdate", "res_id", "module") VALUES (%s, %s, %s, %s, %s)', ('record_id_0003', 'test.search.model', 0, 3, 'module')),
                call('SELECT "a"."id" FROM "test_model" AS "a" WHERE (("a"."name" = %s) AND ("a"."description" = %s))', ('name 2', 'this is a description 2')),
                call('INSERT INTO "ir_model_data" ("name", "model", "noupdate", "res_id", "module") VALUES (%s, %s, %s, %s, %s)', ('record_id_0002', 'test.model', 1, 2, 'module')),
                call('SELECT "a"."res_id" FROM "ir_model_data" AS "a" WHERE (("a"."module" = %s) AND ("a"."name" = %s))', ('module', 'record_id_0002')),
                call('SELECT "a"."id" FROM "test_other_model" AS "a" WHERE (("a"."code" = %s) AND ("a"."test_model_id" = %s))', ('1', 2)),
                call('INSERT INTO "test_other_model" ("code", "test_model_id") VALUES (%s, %s) RETURNING "id"', ('1', 2)),
                call('INSERT INTO "ir_model_data" ("name", "model", "noupdate", "res_id", "module") VALUES (%s, %s, %s, %s, %s)', ('record_id_0004', 'test.other.model', 0, 4, 'module'))
            ]
            expect(cursor.execute.call_args_list).to(contain_exactly(
                *expected_sql
            ))





from oopgrade.version import Version
from expects import *

with description("a Versions class"):
    with context("When comparing"):
        with it("must to known if it's greater from other version"):
            v1 = Version('4.2.4')
            v2 = Version('3.2.1')
            expect(v1).to(be_above(v2))

        with it("must to know if it's greater or equal from other version"):
            v1 = Version('4.2.4')
            v2 = Version('4.2.4')
            expect(v1).to(be_above_or_equal(v2))

        with it("must to know if it's lower than from other version"):
            v1 = Version('4.2.4')
            v2 = Version('3.2.1')
            expect(v2).to(be_below(v1))

        with it("must to know if it's lower or equal than from other version"):
            v1 = Version('4.2.4')
            v2 = Version('4.2.4')
            expect(v2).to(be_below_or_equal(v1))

        with it("must to know if it's equal to other version"):
            v1 = Version('4.2.4')
            v2 = Version('4.2.4')
            expect(v1).to(equal(v2))

    with it('must to bump a major version'):
        v1 = Version('0.1.1')
        v1.bump_major()
        expect(v1).to(equal(Version('1.0.0')))

    with it('must to bump a minor version'):
        v1 = Version('0.1.1')
        v1.bump_minor()
        expect(v1).to(equal(Version('0.2.0')))

    with it('must to bump a patch version'):
        v1 = Version('0.1.1')
        v1.bump_patch()
        expect(v1).to(equal(Version('0.1.2')))

    with it('must to parse and return info as dict'):
        v1 = Version('1.2.3-beta.1+build.4')
        expect(v1.info).to(have_keys(
            major=1,
            minor=2,
            patch=3,
            prerelease='beta.1',
            build='build.4'
        ))

    with it('must be sorted in a list of versions'):
        versions = [Version('2.1.1'), Version('1.0.3'), Version('3.2.4')]
        sorted_version = sorted(versions)
        expect(sorted_version).to(contain_exactly(
            Version('1.0.3'), Version('2.1.1'), Version('3.2.4')
        ))



import semver


class Version(object):
    """Version class using semantic version parsing

    :param version: version string
    """

    def __init__(self, version):
        self.version = version

    @property
    def info(self):
        """Get the information parsing the version.

        :returns: a dict with the keys: major, minor, patch, prerelease and build
        """
        return semver.parse(self.version)

    def bump_major(self):
        """Bump a major version: X from X.Y.Z
        """
        self.version = semver.bump_major(self.version)

    def bump_minor(self):
        """Bump a minor version: Y from X.Y.Z
        """
        self.version = semver.bump_minor(self.version)

    def bump_patch(self):
        """Bump a patch version: Z from X.Y.Z
        """
        self.version = semver.bump_patch(self.version)

    def __gt__(self, other):
        if isinstance(other, basestring):
            other = Version(other)
        return semver.match(self.version, '>{}'.format(other.version))

    def __ge__(self, other):
        if isinstance(other, basestring):
            other = Version(other)
        return semver.match(self.version, '>={}'.format(other.version))

    def __lt__(self, other):
        if isinstance(other, basestring):
            other = Version(other)
        return semver.match(self.version, '<{}'.format(other.version))

    def __le__(self, other):
        if isinstance(other, basestring):
            other = Version(other)
        return semver.match(self.version, '<={}'.format(other.version))

    def __eq__(self, other):
        if isinstance(other, basestring):
            other = Version(other)
        return semver.match(self.version, '=={}'.format(other.version))

    def __unicode__(self):
        return u'v{}'.format(self.version)

    def __str__(self):
        return 'v{}'.format(self.version)

    def __repr__(self):
        return 'v{}'.format(self.version)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

dirname = os.path.dirname(__file__)

setup(
    name='oopgrade',
    version='0.9.0',
    description='Upgrade and migration tools',
    long_description=readme,
    author='GISCE-TI, S.L.',
    author_email='devel@gisce.net',
    url='https://github.com/gisce/oopgrade',
    packages=find_packages(),
    install_requires=[
        'ooquery',
        'semver',
        'lxml',
        'python-sql',
        'tqdm',
        'click',
        'pip',
        'osconf',
    ],
    license='AGPL-3',
    entry_points='''
      [console_scripts]
      oopgrade=oopgrade.cli:oopgrade
    '''
)



# -*- coding: utf-8 -*-
import os
import logging


logger = logging.getLogger('openerp.oopgrade')

MODULE_INSTALLED_STATES = ['installed', 'to upgrade', 'to remove']

__all__ = [
    'load_data',
    'rename_columns',
    'rename_tables',
    'drop_columns',
    'table_exists',
    'column_exists',
    'change_column_type',
    'delete_model_workflow',
    'set_defaults',
    'set_stored_function',
    'update_module_names',
    'add_ir_model_fields',
    'install_modules',
    'get_foreign_keys',
    'get_installed_modules',
    'module_is_installed',
]    


def load_data(cr, module_name, filename, idref=None, mode='init'):
    """
    Load an xml or csv data file from your post script. The usual case for this is the
    occurrence of newly added essential or useful data in the module that is
    marked with "noupdate='1'" and without "forcecreate='1'" so that it will
    not be loaded by the usual upgrade mechanism. Leaving the 'mode' argument to
    its default 'init' will load the data from your migration script.
    
    Theoretically, you could simply load a stock file from the module, but be 
    careful not to reinitialize any data that could have been customized.
    Preferably, select only the newly added items. Copy these to a file
    in your migrations directory and load that file.
    Leave it to the user to actually delete existing resources that are
    marked with 'noupdate' (other named items will be deleted
    automatically).


    :param module_name: the name of the module
    :param filename: the path to the filename, relative to the module \
    directory.
    :param idref: optional hash with ?id mapping cache?
    :param mode: one of 'init', 'update', 'demo'. Always use 'init' for adding new items \
    from files that are marked with 'noupdate'. Defaults to 'init'.

    """
    import tools

    if idref is None:
        idref = {}
    logger.info('%s: loading %s' % (module_name, filename))
    _, ext = os.path.splitext(filename)
    pathname = os.path.join(module_name, filename)
    fp = tools.file_open(pathname)
    try:
        if ext == '.csv':
            noupdate = True
            tools.convert_csv_import(cr, module_name, pathname, fp.read(), idref, mode, noupdate)
        else:
            tools.convert_xml_import(cr, module_name, fp, idref, mode=mode)
    finally:
        fp.close()


def table_exists(cr, table):
    """ Check whether a certain table or view exists """
    cr.execute(
        'SELECT count(relname) FROM pg_class WHERE relname = %s',
        (table,))
    return cr.fetchone()[0] == 1


def rename_columns(cr, column_spec):
    """
    Rename table columns. Typically called in the pre script.

    :param column_spec: a hash with table keys, with lists of tuples as values. \
    Tuples consist of (old_name, new_name).

    """
    for table in column_spec.keys():
        for (old, new) in column_spec[table]:
            logger.info("table %s, column %s: renaming to %s",
                     table, old, new)
            cr.execute('ALTER TABLE "%s" RENAME "%s" TO "%s"' % (table, old, new,))


def rename_tables(cr, table_spec):
    """
    Rename tables. Typically called in the pre script.
    :param column_spec: a list of tuples (old table name, new table name).

    """
    for (old, new) in table_spec:
        logger.info("table %s: renaming to %s",
                    old, new)
        cr.execute('ALTER TABLE "%s" RENAME TO "%s"' % (old, new,))


def rename_models(cr, model_spec):
    """
    Rename models. Typically called in the pre script.
    :param column_spec: a list of tuples (old table name, new table name).
    
    Use case: if a model changes name, but still implements equivalent
    functionality you will want to update references in for instance
    relation fields.

    """
    for (old, new) in model_spec:
        logger.info("model %s: renaming to %s",
                    old, new)
        cr.execute('UPDATE ir_model_fields SET relation = %s '
                   'WHERE relation = %s', (new, old,))


def drop_columns(cr, column_spec):
    """
    Drop columns but perform an additional check if a column exists.
    This covers the case of function fields that may or may not be stored.
    Consider that this may not be obvious: an additional module can govern
    a function fields' store properties.

    :param column_spec: a list of (table, column) tuples
    """
    for (table, column) in column_spec:
        logger.info("table %s: drop column %s",
                    table, column)
        if column_exists(cr, table, column):
            cr.execute('ALTER TABLE "%s" DROP COLUMN "%s"' % 
                       (table, column))
        else:
            logger.warn("table %s: column %s did not exist",
                    table, column)


def add_columns(cr, column_spec):
    """
    Add columns

    :param cr: Database cursor
    :param column_spec: a hash with table keys, with lists of tuples as values.
        Tuples consist of (column name, type).
    """
    for table in column_spec:
        for (column, type_) in column_spec[table]:
            logger.info("table %s: add column %s",
                        table, column)
            if column_exists(cr, table, column):
                logger.warn("table %s: column %s already exists",
                            table, column)
            else:
                cr.execute('ALTER TABLE "%s" ADD COLUMN "%s" %s' %
                           (table, column, type_))


def set_stored_function(cr, obj, fields):
    """
    Init newly created stored functions calling the function and storing them
    to the database.

    ..note:: Use in the post stage

    :param cr: Database cursor
    :param obj: Object
    :param fields: list of fields
    """
    from datetime import datetime

    for k in fields:
        logger.info("storing computed values of fields.function '%s'" % (k,))
        field = obj._columns[k]
        ss = field._symbol_set
        update_query = 'UPDATE "%s" SET "%s"=%s WHERE id=%%s' % (
        obj._table, k, ss[0])
        cr.execute('select id from ' + obj._table)
        ids_lst = map(lambda x: x[0], cr.fetchall())
        logger.info("storing computed values for %s objects" % len(ids_lst))
        start = datetime.now()

        def chunks(l, n):
            """Yield successive n-sized chunks from l."""
            for i in range(0, len(l), n):
                yield l[i:i + n]

        for ids in chunks(ids_lst, 100):
            res = field.get(cr, obj, ids, k, 1, {})
            for key, val in res.items():
                if field._multi:
                    val = val[k]
                # if val is a many2one, just write the ID
                if type(val) == tuple:
                    val = val[0]
                if (val is not False) or (type(val) != bool):
                    cr.execute(update_query, (ss[1](val), key))
        logger.info("stored in {0}".format(datetime.now() - start))


def delete_model_workflow(cr, model):
    """ 
    Forcefully remove active workflows for obsolete models,
    to prevent foreign key issues when the orm deletes the model.
    """
    logged_query(
        cr,
        "DELETE FROM wkf_workitem WHERE act_id in "
        "( SELECT wkf_activity.id "
        "  FROM wkf_activity, wkf "
        "  WHERE wkf_id = wkf.id AND "
        "  wkf.osv = %s"
        ")", (model,))
    logged_query(
        cr,
        "DELETE FROM wkf WHERE osv = %s", (model,))


def set_defaults(cr, pool, default_spec, force=False):
    """
    Set default value. Useful for fields that are newly required. Uses orm, so
    call from the post script.
    
    :param default_spec: a hash with model names as keys. Values are lists of \
    tuples (field, value). None as a value has a special meaning: it assigns \
    the default value. If this value is provided by a function, the function is \
    called as the user that created the resource.
    :param force: overwrite existing values. To be used for assigning a non- \
    default value (presumably in the case of a new column). The ORM assigns \
    the default value as declared in the model in an earlier stage of the \
    process. Beware of issues with resources loaded from new data that \
    actually do require the model's default, in combination with the post \
    script possible being run multiple times.
    """
    from osv import osv

    def write_value(ids, field, value):
        logger.info("model %s, field %s: setting default value of %d resources to %s",
                 model, field, len(ids), unicode(value))
        obj.write(cr, 1, ids, {field: value})

    for model in default_spec.keys():
        obj = pool.get(model)
        if not obj:
            raise osv.except_osv("Migration: error setting default, no such model: %s" % model, "")

    for field, value in default_spec[model]:
        domain = not force and [(field, '=', False)] or []
        ids = obj.search(cr, 1, domain)
        if not ids:
            continue
        if value is None:
            # Set the value by calling the _defaults of the object.
            # Typically used for company_id on various models, and in that
            # case the result depends on the user associated with the object.
            # We retrieve create_uid for this purpose and need to call the _defaults
            # function per resource. Otherwise, write all resources at once.
            if field in obj._defaults:
                if not callable(obj._defaults[field]):
                    write_value(ids, field, obj._defaults[field])
                else:
                    # existence users is covered by foreign keys, so this is not needed
                    # cr.execute("SELECT %s.id, res_users.id FROM %s LEFT OUTER JOIN res_users ON (%s.create_uid = res_users.id) WHERE %s.id IN %s" %
                    #                     (obj._table, obj._table, obj._table, obj._table, tuple(ids),))
                    cr.execute("SELECT id, COALESCE(create_uid, 1) FROM %s " % obj._table + "WHERE id in %s", (tuple(ids),))
                    fetchdict = dict(cr.fetchall())
                    for id in ids:
                        write_value([id], field, obj._defaults[field](obj, cr, fetchdict.get(id, 1), None))
                        if id not in fetchdict:
                            logger.info("model %s, field %s, id %d: no create_uid defined or user does not exist anymore",
                                     model, field, id)
            else:
                error = ("OpenUpgrade: error setting default, field %s with "
                         "None default value not in %s' _defaults" % (
                        field, model))
                logger.error(error)
                # this exeption seems to get lost in a higher up try block
                osv.except_osv("OpenUpgrade", error)
        else:
            write_value(ids, field, value)


def logged_query(cr, query, args=None):
    if args is None:
        args = []
    res = cr.execute(query, args)
    logger.debug('Running %s', query)
    if not res:
        query = query % args
        logger.warn('No rows affected for query "%s"', query)
    return res


def column_exists(cr, table, column):
    """
    Check whether a certain column exists

    :param cr: Database cursor
    :param table: Table name
    :type table: str or unicode
    :param column: Column
    :type column: str or unicode
    :return: True if the column exists
    :rtype: bool
    """
    cr.execute(
        'SELECT count(attname) FROM pg_attribute '
        'WHERE attrelid = '
        '( SELECT oid FROM pg_class WHERE relname = %s ) '
        'AND attname = %s',
        (table, column));
    return cr.fetchone()[0] == 1


def change_column_type(cursor, column_spec):
    """
    :param cr: Cursor
    :param colum_spec: a hash with table keys, with lists of tuples as values.
        Tuples consist of (column name, new_def). new_def as in
        posgresql_language
    :return: execute result
    """
    for table, spec in column_spec.items():
        for column, new_def in spec:
            logged_query(
                cursor,
                'ALTER TABLE %s ALTER COLUMN %s TYPE %s' % (
                    table, column, new_def)
            )
    return True


def update_module_names(cr, namespec):
    """
    Deal with changed module names of certified modules
    in order to prevent  'certificate not unique' error,
    as well as updating the module reference in the
    XML id.
    
    :param namespec: tuple of (old name, new name)
    """
    for (old_name, new_name) in namespec:
        query = ("UPDATE ir_module_module SET name = %s "
                 "WHERE name = %s")
        logged_query(cr, query, (new_name, old_name))
        query = ("UPDATE ir_model_data SET module = %s "
                 "WHERE module = %s ")
        logged_query(cr, query, (new_name, old_name))


def add_ir_model_fields(cr, columnspec):
    """
    Typically, new columns on ir_model_fields need to be added in a very
    early stage in the upgrade process of the base module, in raw sql
    as they need to be in place before any model gets initialized.
    Do not use for fields with additional SQL constraints, such as a
    reference to another table or the cascade constraint, but craft your
    own statement taking them into account.
    
    :param columnspec: tuple of (column name, column type)
    """
    for column in columnspec:
        query = 'ALTER TABLE ir_model_fields ADD COLUMN %s %s' % (
            column)
        logged_query(cr, query, [])


def install_modules(cursor, *modules):
    """Installs a module.

    :param cr: Cursor database
    :param module: The module to install
    """
    import pooler

    uid = 1
    mod_obj = pooler.get_pool(cursor.dbname).get('ir.module.module')
    mod_obj.update_list(cursor, uid)
    search_params = [('name', 'in', modules), ('state', '=', 'uninstalled')]
    mod_ids = mod_obj.search(cursor, uid, search_params)
    mod_obj.button_install(cursor, uid, mod_ids)
    return True


def get_foreign_keys(cursor, table):
    """Get all the foreign keys from the given table

    Returns a dict with column_name as a key and the following keys:
      - constraint_name
      - table_name
      - column_name
      - foreign_table_name
      - foreign_column_name

    :param cursor: Database cursor
    :param table: Table name to get the foreign keys
    :return: dict
    """
    cursor.execute(
        "SELECT"
        " tc.constraint_name, tc.table_name, kcu.column_name,"
        " ccu.table_name AS foreign_table_name,"
        " ccu.column_name AS foreign_column_name"
        " FROM "
        "   information_schema.table_constraints AS tc"
        " JOIN information_schema.key_column_usage AS kcu"
        "   ON tc.constraint_name = kcu.constraint_name"
        " JOIN information_schema.constraint_column_usage AS ccu"
        "   ON ccu.constraint_name = tc.constraint_name"
        " WHERE constraint_type = 'FOREIGN KEY' AND tc.table_name=%s",
        (table,)
    )
    res = {}
    for fk in cursor.dictfetchall():
        res[fk['column_name']] = fk.copy()
    return res


def get_installed_modules(cursor):
    cursor.execute(
        "SELECT"
        " name "
        "FROM "
        "  ir_module_module "
        "WHERE state = 'installed'"
    )
    return [x[0] for x in cursor.fetchall()]

  
def module_is_installed(cursor, module_name):
    """Test if modules is installed.

    :param cr: Cursor database
    :param module_name: The module name
    """
    import pooler

    uid = 1
    mod_obj = pooler.get_pool(cursor.dbname).get('ir.module.module')
    search_params = [('name', '=', module_name),
                     ('state', 'in', MODULE_INSTALLED_STATES)]
    mod_ids = mod_obj.search(cursor, uid, search_params)
    return len(mod_ids) > 0



# -*- coding: utf-8 -*-
from oopgrade import table_exists


def log_xml_id(cr, module, xml_id):
    """
    Log xml_ids at load time in the records table.
    Called from openerp/tools/convert.py:xml_import._test_xml_id()

    # Catcha's
    - The module needs to be loaded with 'init', or the calling method
    won't be called. This can be brought about by installing the
    module or updating the 'state' field of the module to 'to install'
    or call the server with '--init <module>' and the database argument.
    
    - Do you get the right results immediately when installing the module?
    No, sorry. This method retrieves the model from the ir_model_table, but when
    the xml id is encountered for the first time, this method is called
    before the item is present in this table. Therefore, you will not
    get any meaningful results until the *second* time that you 'init'
    the module.

    - The good news is that the openupgrade_records module that comes
    with this distribution allows you to deal with all of this with
    one click on the menu item Settings -> Customizations ->
    Database Structure -> OpenUpgrade -> Generate Records

    - You cannot reinitialize the modules in your production database
    and expect to keep working on it happily ever after. Do not perform
    this routine on your production database.

    :param module: The module that contains the xml_id
    :param xml_id: the xml_id, with or without 'module.' prefix
    """
    if not table_exists(cr, 'openupgrade_record'):
        return
    if not '.' in xml_id:
        xml_id = '%s.%s' % (module, xml_id)
    cr.execute(
        "SELECT model FROM ir_model_data "
        "WHERE module = %s AND name = %s",
        xml_id.split('.'))
    record = cr.fetchone()
    if not record:
        #print "Cannot find xml_id %s" % xml_id
        return
    else:
        cr.execute(
            "SELECT id FROM openupgrade_record "
            "WHERE module=%s AND model=%s AND name=%s AND type=%s",
            (module, record[0], xml_id, 'xmlid'))
        if not cr.fetchone():
            cr.execute(
                "INSERT INTO openupgrade_record "
                "(module, model, name, type) values(%s, %s, %s, %s)",
                (module, record[0], xml_id, 'xmlid'))

